#!/usr/bin/perl -s

#
# sub-sampling photomosaic - Jim Bumgardner
#

use Image::Magick;


require 'samples.ph';
# this is an array with data about each thumbnail - computed when I download the images, 
#       it reads as follows:
# @images = (
#   {id=>3875606,w=>100,h=>74,r=>81,g=>67,b=>63,l=>70.9924,owner=>'99083727@N00',secret=>'9a78d53bf4',server=>2},
#   {id=>3875586,w=>100,h=>75,r=>102,g=>82,b=>72,l=>87.352,owner=>'99083727@N00',secret=>'23fbb36ae0',server=>3},
#      etc...
# );
#
# -cellSize=N   defaults to 100
# -noswap       skip swap pass

$reso = 3 if !$reso;  # number of subsamples per photo = reso x reso
$reso2 = $reso**2;
$cellSize = 100 if !$cellSize;
$lumErr = 20 if !$lumErr; # $max allowed luminance deviation
$basepic = 'mona2.png' if !$basepic;

print "reso = $reso\n";
print "cellSize = $cellSize\n";
print "lumErr = $lumErr\n";
print "basepic = $basepic\n";

# perform sub-sampling to arbitrary resolution here...
print "Sampling...\n";
$n = 0;
for $photo (@images)
{
	  $fnam = sprintf "snaps/%d_%s_t.jpg", $photo->{id}, $photo->{secret};
	  my $image = Image::Magick->new;
		$err = $image->Read($fnam);
		warn "$err\n" if $err;
		# crop to square
	  my ($w, $h) = $image->Get('width', 'height');
	  if ($w < $h) {
	  	$image->Crop(width=>$w, height=>$w, x=>0, y=>($h-$w)/2);
	  }
	  elsif ($w > $h) {
	  	$image->Crop(width=>$h, height=>$h, y=>0, x=>($w-$h)/2);
	  }
		$x = $image->Resize(width=>$reso, height=>$reso);
		my @pix = ();
		for $py (0..$reso-1)
		{
			for $px (0..$reso-1)
			{
					($r,$g,$b) = split /,/,$image->Get("pixel[$px,$py]");
					push @pix, [$r,$g,$b];
			}
		}
		$photo->{pix} = \@pix;
		undef $image;
		print "$n...\n" if ++$n % 500 == 0;
}



$baseimg = Image::Magick->new;
$err = $baseimg->Read($basepic);
warn "$err" if "$err";

my ($w, $h) = $baseimg->Get('width', 'height');

my $aspect = $h/$w;

my $hcells = sqrt(@images / $aspect);
my $vcells = $hcells * $aspect;
$hcells = int($hcells);
$vcells = int($vcells);

print "H X V = $hcells x $vcells\n";

# this version is sized with 1 pixel per tile
$baseimg->Resize(width=>$hcells, height=>$vcells);

# this version is sized with resoXreso pixels per tile
# used for sub-sample matching
$baseimg2 = Image::Magick->new;
$err = $baseimg2->Read($basepic);
warn "$err" if "$err";
$baseimg2->Resize(width=>$hcells*$reso, height=>$vcells*$reso);

%imageKey = ();
for $photo (@images)
{
	$imageKey{$photo->{id}} = $photo;
}


$n = @images;
$m = $hcells * $vcells;

@images = @images[$n-$m .. $n-1];

$n = @images;

print "Number of images in square: $hcells x $vcells = $n\n";


$cx = $hcells/2;
$cy = $vcells/2;

@cells = ();

# in this pass, I subsample the image and store pixel information into each cell
#
$i = 0;
for $y (0..$vcells-1)
{
  for $x (0..$hcells-1)
  {
		$p = $baseimg->Get("pixel[$x,$y]");
		my ($r,$g,$b) = split /,/,$p;
		my $l = 0.3086*$r + 0.6094*$g + 0.0820*$b; # Haeberli

		my $cl = Graphics::ColorObject->newRGB($r,$g,$b);
	  my ($h, $s, $v) = $cl->asHSV;

		my ($x0,$y0) = ($x*$reso, $y*$reso);
		
		my $cell = {i=>$i, x=>$x, y=>$y, r=>$r, g=>$g, b=>$b, l=>$l, h=>$h};

		my @pix = ();
		
		for $py (0..$reso-1)
		{
			for $px (0..$reso-1)
			{
			  my ($x1,$y1) = ($x0+$px, $y0+$py);
				my ($r1,$g1,$b1) = split /,/,$baseimg2->Get("pixel[$x1,$y1]");
				push @pix, [$r1, $g1, $b1];
			}
		}
		$cell->{pix} = \@pix;
	  push @cells, $cell;
	  $i++;
  }
}

#
# sort cells by luminance

@scells = sort { $a->{l} <=> $b->{l}; } @cells;

use Graphics::ColorObject;

# sort tiles by lum
@images = sort {$a->{l} <=> $b->{l}} @images;


# do initial tile placement based on luminence sort
#
$i = 0;
for $photo (@images)
{
  my ($r, $g, $b, $l) = ($photo->{r}, $photo->{g}, $photo->{b}, $photo->{l});
	my $cl = Graphics::ColorObject->newRGB($r,$g,$b);
  my ($h, $s, $v) = $cl->asHSV;
  $photo->{h} = $h; # hue
  $photo->{s} = $s; # saturation
	$photo->{i} = $scells[$i]->{i};
	$i++;
}


print "Checking Swaps...";

# do the hue-swap passes - this is the most time consuming part

if (!$noswap)
{
	$startswap = time();
	# sort by descending saturation- prioritizes bright colors
	@images = sort {$b->{s} <=> $a->{s}} @images;

	$minSwapPasses = 10;
	$swapPasses = 0;
	do {
		$nbrSwaps = 0;
		for $i (1..$m-1)
		{
			my $c1 = $images[$i];
			for $j (0..$i-1)
			{
				my $c2 = $images[$j];
				if (abs($c2->{l} - $c1->{l}) < $lumErr)
				{
					# compute error from c1 to position c1 to dhc1
					$cell1 = $cells[$c1->{i}];
					$cell2 = $cells[$c2->{i}];
					my $dhi1 = CumDiff($c1, $cell1);
					my $dhi2 = CumDiff($c1, $cell2);
					my $dhj1 = CumDiff($c2, $cell2);
					my $dhj2 = CumDiff($c2, $cell1);
					
					if (  $dhi1 > $dhi2 && $dhj1 > $dhj2   )
					{
						# swap cellIndices
						my $tmp = $c2->{i};
						$c2->{i} = $c1->{i};
						$c1->{i} = $tmp;
						$nbrSwaps++;
					}
				}
			}
		}
		++$swapPasses;
		print "Swap Pass $swapPasses, $nbrSwaps swaps...\n";
	} while $nbrSwaps > $minSwapPasses;
	$endswap = time();
	printf "Swap time: %d secs\n", $endswap - $startswap;
}


#
# Output final image
#

my ($width, $height) = ($cellSize*$hcells, $cellSize*$vcells);

$mosaic = Image::Magick->new;
$mosaic->Set(size=>"${width}x${height}"); # worked okay at 10000x10000
$err = $mosaic->Read('xc:black');
warn "$err" if "$err";

print "Outputing Image $width x $height...\n";

$i = 0;
foreach $imgdat (@images)
{
  my ($x, $y) = ($cells[$imgdat->{i}]->{x}, $cells[$imgdat->{i}]->{y});

	$fnam = sprintf "%d_%s.jpg", $imgdat->{id}, $imgdat->{secret};
	$fnam =~ s~\.jpg~_t.jpg~ if $cellSize <= 100;
	$fnam = "snaps/$fnam";

  $img = Image::Magick->new;

  $err = $img->Read($fnam);	
  warn "$err" if "$err";

  my ($w, $h) = $img->Get('width', 'height');
  if ($w < $h) {
  	$img->Crop(width=>$w, height=>$w, x=>0, y=>($h-$w)/2);
  }
  elsif ($w > $h) {
  	$img->Crop(width=>$h, height=>$h, y=>0, x=>($w-$h)/2);
  }
	$img->Resize(width=>$cellSize, height=>$cellSize);

  $err = $mosaic->Composite(image=>$img, x=>$x*$cellSize, y=>$y*$cellSize);
  warn "$err" if "$err";

  undef $img;
  print "$i...\n" if (++$i % 500 == 0);
}

print "Saving JPEG...\n";

$mosaic->Set(quality=>90);
$mosaic->Write("mosaicx_${reso}x${reso}_e${lumErr}_${hcells}_x_${vcells}.jpg");


#
# subroutine to compute squared color distance from tile to cell
#
sub CumDiff($$)
{
  my ($img, $cell) = @_;
  
  my $sum = 0;
  
	my ($pix1,$pix2) = ($cell->{pix}, $pix2 = $img->{pix});

	for my $i (0..$reso2-1)
	{
		my ($p1, $p2) = ($pix1->[$i], $pix2->[$i]);
		$sum += ($p1->[0] - $p2->[0])**2 +
		        ($p1->[1] - $p2->[1])**2 +
		        ($p1->[2] - $p2->[2])**2;
	}

  return $sum;
}